'''
本模块用于数据预处理

This module is used for data preproccessing
'''
import numpy as np
from maysics.utils import e_distances


def shuffle(*arg):
    '''
    以相同方法打乱多个序列或打乱一个序列
    
    返回值：一个ndarray
    
    
    shuffle multiple sequences in the same way or shuffle a sequences
    
    return: a ndarray
    '''
    state = np.random.get_state()
    a_new_list = []
    for li in arg:
        np.random.set_state(state)
        np.random.shuffle(li)
        a_new_list.append(li)
    return np.array(a_new_list)


def data_split(data, targets, train_size=None, test_size=None, shuffle=True, random_state=None):
    '''
    分离数据
    
    参数
    ----
    data：数据
    targets：指标
    train_size：浮点数类型，可选，训练集占总数据量的比，取值范围为(0, 1]，默认为0.75
    test_size：浮点数类型，可选，测试集占总数据量的比，取值范围为[0, 1)，当train_size被定义时，该参数无效
    shuffle：布尔类型，可选，True表示打乱数据，False表示不打乱数据，默认为True
    random_state：整型，可选，随机种子
    
    
    split the data
    
    Parameters
    ----------
    data: data
    targets: targets
    train_size: float, callable, ratio of training set to total data, value range is (0, 1], default=0.75
    test_size: float, callable, ratio of test set to total data, value range is [0, 1)
    shuffle: bool, callable, 'True' will shuffle the data, 'False' will not, default = True
    random_state: int, callable, random seed
    '''
    if not (train_size or test_size):
        train_size = 0.75
    elif test_size:
        train_size = 1 - test_size
    
    if train_size <= 0 or train_size > 1:
        raise Exception("'train_size' should be in (0, 1], 'test_size' should be in [0, 1)")
    
    if shuffle:
        np.random.seed(random_state)
        
        state = np.random.get_state()
        np.random.shuffle(data)
        np.random.set_state(state)
        np.random.shuffle(targets)
        
    num_of_data = len(data)
    train_data = data[:int(num_of_data * train_size)]
    train_target = targets[:int(num_of_data * train_size)]
    validation_data = data[int(num_of_data * train_size):]
    validation_target = targets[int(num_of_data * train_size):]
    
    return train_data, train_target, validation_data, validation_target


def kfold(data, targets, n, k=5):
    '''
    参数
    ----
    data：数据
    targets：指标
    n：整型，表示将第n折作为验证集，从0开始
    k：整型，可选，k折验证的折叠数，默认k=5
    
    返回值：训练集和验证集的元组
    
    
    Parameters
    ----------
    data: data
    targets: targets
    n: int, take the nth part as validation set, starting from 0
    k: int, callable, the number of k-fold, default = 5
    
    return: the tuple of training set and validation set
    '''
    num_validation_samples = len(data) // k
    
    validation_data = data[num_validation_samples * n:
                           num_validation_samples * (n + 1)]
    validation_targets = targets[num_validation_samples * n:
                                 num_validation_samples * (n + 1)]
    
    train_data = np.concatenate([data[: num_validation_samples * n],
                                 data[num_validation_samples * (n + 1):]])
    train_targets = np.concatenate([targets[: num_validation_samples * n],
                                    targets[num_validation_samples * (n + 1):]])
    
    return train_data, train_targets, validation_data, validation_targets


def standard(data, index=None, mean=True, var=True):
    '''
    标准化数据
    z = (x - u) / s
    z：新数据；  x：原数据；  u：均值；  s：方差
    如果某一列数据完全相同（即方差s=0），则该列数据全部归零
    
    参数
    ----
    data：2-D的ndarray数据
    index：列表形式，可选，需要进行标准化的列的索引，默认为全部
    mean：布尔类型，可选，是否将均值调整为0
    var：布尔类型，可选，是否将方差调整为1
    
    
    Standardize data
    z = (x - u) / s
    z: new data;  x: origin data;  u: mean value;  s: variance
    if data in one column are the same(s=0), data in this column will be turned to 0
    
    Parameters
    ----------
    data: 2-D ndarray
    index: list, callable, index of columns need to be standardized, defalut to all
    mean: bool, callable, if adjust the mean value to 0
    var: bool, callable, if adjust the variance to 0
    '''
    data=np.array(data, dtype=np.float)
    
    if index:
        if mean:
            mean = data[:, index].mean(axis=0)
        else:
            mean = 0.0
        data[:, index] -= mean
        
        if var:
            std = data[:, index].std(axis=0)
            std_zero_indices = np.nonzero(std == 0)
            std[std==0] = 1.0
            data[:, index] /= std
            if list(std_zero_indices[0]):
                for i in std_zero_indices[0]:
                    data[:, index][:, i] *= 0
    else:
        if mean:
            mean = data.mean(axis=0)
        else:
            mean = 0.0
        data -= mean
        
        if var:
            std = data.std(axis=0)
            std_zero_indices = np.nonzero(std == 0)
            std[std==0] = 1.0
            data /= std
            if list(std_zero_indices[0]):
                for i in std_zero_indices[0]:
                    data[:, i] *= 0
    
    return data


def minmax(data, index=None, feature_range=(0, 1)):
    '''
    归一化数据
    X_std = (X - X.min(axis=0)) / (X.max(axis=0) - X.min(axis=0))
    X_new = X_std * (feature_range[1] - feature_range[0]) + feature_range[0]
    
    参数
    ----
    data：2-D的ndarray数据
    index：列表形式，可选，需要进行标准化的列的索引，默认为全部
    feature_range：元组形式，可选，需要转换的范围，默认为(0, 1)
    
    
    Normalized data
    X_std = (X - X.min(axis=0)) / (X.max(axis=0) - X.min(axis=0))
    X_new = X_std * (feature_range[1] - feature_range[0]) + feature_range[0]
    
    Parameters
    ----------
    data: 2-D的ndarray数据
    index: list, callable, index of columns need to be standardized, defalut to all
    featur_range: tuple, callabel, final range of transformed data
    '''
    data=np.array(data, dtype=np.float)
    
    if index:
        min_data = data[:, index].min(axis=0)
        length = data[:, index].max(axis=0) - min_data
        data[:, index] = (data[:, index] - min_data) / length
        data[:, index] = data[:, index] * (feature_range[1] - feature_range[0]) + feature_range[0]
    
    else:
        min_data = data.min(axis=0)
        length = data.max(axis=0) - min_data
        data = (data - min_data) / length
        data = data * (feature_range[1] - feature_range[0]) + feature_range[0]
    
    return data


def normalizer(data, index=None):
    '''
    使每个数据的模为1
    
    参数
    ----
    data：2-D的ndarray数据
    index：列表形式，可选，需要进行标准化的列的索引，默认为全部
    
    
    Making the moduli of data equal 1
    
    Parameters
    ----------
    data: 2-D的ndarray数据
    index: list, callable, index of columns need to be standardized, defalut to all
    '''
    data = np.array(data, dtype=np.float)
    
    if index:
        distance_list = e_distances(data[:, index])
        distance_list[distance_list == 0] = 1
        print(distance_list)
        data[:, index] /= distance_list
    
    else:
        distance_list = e_distances(data)
        distance_list[distance_list == 0] = 1
        print(distance_list)
        data /= distance_list
    
    return data


'''
本库用于科学计算和快速建模

maysics主要包括九个模块：

1、algorithm 封装了几种模拟方法，用于简易模拟；
2、calculus 封装了部分常见的算符算子和积分方法，辅助数学运算；
3、constant 储存了部分常数；
4、model_selection 用于评估和选择模型；
5、models 封装了几种常用的数学物理定律、方程、模型以便快速构建数理模型；
6、preprocess 用于数据预处理；
7、stats 用于统计分析；
8、transformation 储存了常用的坐标转换及其他数学变换；
9、utils 是额外工具箱。


This package is used for scientific calculating and fast modeling.

maysics includes nine modules:

1. "algorithm" packages several simulation methods for simple simulation;
2. "calculus" packages some common operators and integration method to assist in mathematical operations;
3. "constant" contents some usual constants;
4. "model_selection" used for estimating and selecting model;
5. "models" packages several commonly used laws, equations and models of mathematical physics for fast modeling;
6. "preprocess" is used for data preproccessing;
7. "stats" is uesd for statistical analysis;
8. "transformation" stores common coordinate transformations and other mathematical transformations;
9. "utils" is extra Utils.

maysics-|- __init__
        |
        |- algorithm -------|- GA
        |                   |- MC
        |                   |- SA
        |
        |- calculus --------|- Del
        |                   |- Inte
        |
        |- constant --------|- LP
        |
        |- model_selection -|- Error
        |                   |- Estimate
        |                   |- Search
        |                   |- Sense
        |
        |- models ----------|- ED
        |                   |- Fouriers_law
        |                   |- Leslie
        |                   |- Logistic
        |                   |- MVD_law
        |                   |- Plancks_law
        |
        |- preprocess
        |
        |- stats -----------|- DFT
        |
        |- transformation
        |
        |- utils -----------|- rc
'''

import numpy as np
import pickle
from maysics import algorithm, calculus, constant, model_selection, models, preprocess, stats, transformation, utils


def arr(f):
    '''
    将矢量函数的输出形式统一为ndarray
    
    返回值：更改输出格式后的函数
    
    
    transform the output of vector function as ndarray
    
    return: function after changing the output format
    '''
    def obj(x):
        func = f(x)
        try:
            return np.array(func)
        except:
            return func
    return obj


def add(*arg):
    '''
    实现函数与同型函数、函数与数之间的加法
    要求作用函数若输出列表，必须是ndarray格式
    
    返回值：相加后的新函数
    
    
    addition between function and function or function and number
    if output of the function is list, it requires ndarray
    
    return: new function after addition
    '''
    def obj(x):
        list = []
        for i in range(len(arg)):
            if type(arg[i]).__name__ == 'function':
                list.append(arg[i](x))
            else:
                list.append(arg[i])
        return sum(list)
    return obj


def mul(*arg):
    '''
    实现函数与同型函数、函数与数之间的乘法
    要求作用函数若输出列表，必须是ndarray格式
    
    返回值：相乘后的新函数
    
    
    multiplication between function and function or function and number
    if output of the function is list, it requires ndarray
    
    return: new function after multiplication
    '''
    def obj(x):
        result = 1
        for i in range(len(arg)):
            if type(arg[i]).__name__ == 'function':
                result *= arg[i](x)
            else:
                result *= arg[i]
        return result
    return obj


def sub(minuend, subtrahend):
    '''
    实现函数与同型函数、函数与数之间的减法
    要求作用函数若输出列表，必须是ndarray格式
    minuend：被减数
    subtrahend：减数
    
    返回值：相乘后的新函数
    
    
    subtraction between function and function or function and number
    if output of the function is list, it requires ndarray
    minuend: minuend
    subtrahend: subtrahend
    
    return: new function after subtraction
    '''
    def obj(x):
        if type(minuend).__name__ == 'function':
            result_of_minuend = minuend(x)
        else:
            result_of_minuend = minuend
        if type(subtrahend).__name__ == 'function':
            result_of_subtrahend = subtrahend(x)
        else:
            result_of_subtrahend = subtrahend
        
        result = result_of_minuend - result_of_subtrahend
        return result
    return obj


def divi(dividend, divisor):
    '''
    实现函数与同型函数、函数与数之间的除法
    要求作用函数若输出列表，必须是ndarray格式
    
    参数
    ----
    dividend：被除数
    divisor：除数
    
    返回值：相乘后的新函数
    
    
    Parameters
    ----------
    division between function and function or function and number
    if output of the function is list, it requires ndarray
    dividend: dividend
    divisor: divisor
    
    return: new function after division
    '''
    def obj(x):
        if type(dividend).__name__ == 'function':
            result_of_dividend = dividend(x)
        else:
            result_of_dividend = dividend
        if type(divisor).__name__ == 'function':
            result_of_divisor = divisor(x)
        else:
            result_of_divisor = divisor
        result = result_of_dividend - result_of_divisor
        return result
    return obj


def save(data, filename):
    '''
    保存为.pkl或.npy文件
    
    参数
    ----
    data：需要保存的数据
    filename：字符串类型，文件名
    
    
    Save as .pkl or .npy file
    
    Parameters
    ----------
    data: data
    filename: str, file name
    '''
    if filename[-4:] == '.pkl':
        with open(filename, 'wb') as file:
            pickle.dump(data, file)
    
    elif filename[-4:] == '.npy':
        np.save(filename, data)
    
    else:
        raise Exception("Suffix of filename must be '.pkl' or '.npy'.")


def load(filename):
    '''
    载入.pkl或.npy文件
    
    参数
    ----
    filename：字符串类型，文件名
    
    
    Load .pkl or .npy file
    
    Parameter
    ---------
    filename: str, file name
    '''
    if filename[-4:] == '.pkl':
        with open(filename, 'rb') as file:
            data = pickle.load(file)
        
        return data
    
    elif filename[-4:] == '.npy':
        return np.load(filename, allow_pickle=True)
    
    else:
        raise Exception("Suffix of filename must be '.pkl' or '.npy'.")


'''
本模块用于评估和选择模型

This module is ued for estimating and selecting models
'''

import numpy as np
from matplotlib import pyplot as plt
from maysics.preprocess import data_split


class Estimate():
    '''
    参数
    ----
    data：ndarray形式，自变量
    targets：ndarray形式，因变量
    verbose：整型，可选，可以是0、1、2，默认为0
    
    
    Parameters
    ----------
    data: ndarray
    targets: ndarray
    verbose: int, callable, can be 0, 1, 2, default = 0
    '''
    def __init__(self, data, targets, verbose=0):
        self.data = data
        self.targets = targets
        self.__verbose = verbose
    
    
    def __part_of_kfold(self, num_validation_samples, num_part):
        validation_data = self.data[num_validation_samples * num_part:
                                    num_validation_samples * (num_part + 1)]
        validation_targets = self.targets[num_validation_samples * num_part:
                                          num_validation_samples * (num_part + 1)]
        
        train_data = np.concatenate([self.data[: num_validation_samples * num_part],
                                     self.data[num_validation_samples * (num_part + 1):]])
        train_targets = np.concatenate([self.targets[: num_validation_samples * num_part],
                                        self.targets[num_validation_samples * (num_part + 1):]])
        
        return train_data, train_targets, validation_data, validation_targets


    def keras_normal(self, build_model, num_epochs, batch_size, train_size=None, val_size=None, param=None, shuffle=True, random_state=None):
        ''' 
        对keras模型或与keras模型相同架构的模型进行数据分割和训练
        
        参数
        ----
        build_model：模型函数，需要返回一个model
        num_epochs：整型，迭代次数
        batch_size：整型，批量大小
        train_size：浮点数类型，可选，训练集占总数据量的比，取值范围为(0, 1]，默认为0.75
        val_size：浮点数类型，可选，测试集占总数据量的比，取值范围为[0, 1)，当train_size被定义时，该参数无效
        shuffle：布尔类型，可选，True表示打乱数据，False表示不打乱数据，默认为True
        param：字典类型，可选，当build_model函数有参数时，需输入以参数名为键，参数值为值的字典
        random_state：整型，可选，随机种子
        
        
        Spliting data and training for keras models and other similar models
        
        Parameters
        ----------
        build_model: the function to generate keras-model, which should return a model
        num_epochs: int, the number of epochs
        batch_size: int, the size of batch
        train_size: float, callable, ratio of training set to total data, value range is (0, 1], default=0.75
        val_size: float, callable, ratio of test set to total data, value range is [0, 1)
        shuffle: bool, callable, 'True' will shuffle the data, 'False' will not, default = True
        param: dict, callable, when build_model() has non-default parameters, param needs to be input a dictionary with parm_name as key and param_value as value
        random_state: int, callable, random seed
        '''
        self.__num_epochs = num_epochs
        np.random.seed(random_state)
        
        if shuffle:
            local_data = self.data
            local_targets = self.targets
            state = np.random.get_state()
            np.random.shuffle(local_data)
            np.random.set_state(state)
            np.random.shuffle(local_targets)
        
        if not (train_size or val_size):
            train_size = 0.75
        elif val_size:
            train_size = 1 - val_size
    
        if train_size <= 0 or train_size > 1:
            raise Exception("'train_size' should be in (0, 1], 'test_size' should be in [0, 1)")
        
        num_of_data = int(len(local_data) * train_size)
        train_data = local_data[:num_of_data]
        train_targets = local_targets[:num_of_data]
        validation_data = local_data[num_of_data:]
        validation_targets = local_targets[num_of_data * train_size:]
                
        if param:
            model = build_model(**param)
        else:
            model = build_model()
        
        history = model.fit(train_data, train_targets,
                            validation_data=(validation_data, validation_targets),
                            epochs=num_epochs, batch_size=batch_size, verbose=self.__verbose)
        self.history = history.history
    
    
    def keras_rekfold(self, build_model, num_epochs, batch_size, k=5, n=3, shuffle=True, param=None, random_state=None):
        '''
        对keras模型或与keras模型相同架构的模型进行重复的k折验证
        
        参数
        ----
        build_model：模型函数，需要返回一个model
        num_epochs：整型，迭代次数
        batch_size：整型，批量大小
        k：整型，可选，k折验证的折叠数，默认为5
        n：整型，可选，k折验证的重复次数，默认为3
        shuffle：布尔类型，可选，True表示打乱数据，False表示不打乱数据，默认为True
        param：字典类型，可选，当build_model函数有参数时，需输入以参数名为键，参数值为值的字典
        random_state：整型，可选，随机种子
        
        
        Repeated k-fold verification for keras models and other similar models
        
        Parameters
        ----------
        build_model: the function to generate keras-model, which should return a model
        num_epochs: int, the number of epochs
        batch_size: int, the size of batch
        k: int, callable, the number of k-fold, default = 5
        n: int, callable, the number of repetition of k-fold, default = 3
        shuffle: bool, callable, 'True' will shuffle the data, 'False' will not, default = True
        param: dict, callable, when build_model() has non-default parameters, param needs to be input a dictionary with parm_name as key and param_value as value
        random_state: int, callable, random seed
        '''
        self.__num_epochs = num_epochs
        num_validation_samples = len(self.data) // k
        np.random.seed(random_state)
        
        for i in range(n):
            if shuffle:
                local_data = self.data
                local_targets = self.targets
                state = np.random.get_state()
                np.random.shuffle(local_data)
                np.random.set_state(state)
                np.random.shuffle(local_targets)
            for j in range(k):
                train_data, train_targets,\
                validation_data, validation_targets = Estimate.__part_of_kfold(
                    self, num_validation_samples=num_validation_samples, num_part=j)
                
                if param:
                    model = build_model(**param)
                else:
                    model = build_model()
                history = model.fit(train_data, train_targets,
                                    validation_data=(validation_data, validation_targets),
                                    epochs=num_epochs, batch_size=batch_size, verbose=self.__verbose)
                if i == 0 and j == 0:
                    new_keys = list(history.history.keys())
                    new_values_of_history = np.zeros_like(list(history.history.values()))
                
                new_values_of_history += np.array(list(history.history.values()))
        
        new_values_of_history /= (n * k)
        history_dict = dict(zip(new_keys, new_values_of_history))
        self.history = history_dict
    
    
    def __img_process(self, epochs_range, metrics_list):
        epochs_range_list = np.arange(epochs_range[0], epochs_range[1]+1)
        
        if metrics_list == 'all':
            metrics_list = list(self.history.keys())
            pic_num = int(len(metrics_list) / 2)
            metrics_list = metrics_list[pic_num :]
        else:
            pic_num = len(metrics_list)
        fig = plt.figure(figsize=(6, 4 * pic_num))
        cal_num = 1
        for metric in metrics_list:
            val_range_list = self.history['val_'+metric][epochs_range[0]-1: epochs_range[1]]
            range_list = self.history[metric][epochs_range[0]-1: epochs_range[1]]
            
            ax = fig.add_subplot(pic_num, 1, cal_num)
            ax.plot(epochs_range_list, val_range_list, label='val_'+metric)
            ax.plot(epochs_range_list, range_list, label=metric)
            ax.legend()
            ax.set_title(metric)
            
            cal_num += 1
        
    
    
    def show(self, epochs_range=None, metrics_list='all'):
        '''
        作指标值-迭代次数图像并显示
        
        参数
        ----
        epochs_range：列表类型，可选，定义图像的绘制范围，默认全部画出
        metrics_list：列表类型，指定绘图的指标函数列表，默认为全部指标
        
        
        Display the metrics - epochs image
        
        Parameters
        ----------
        epochs_range: list, callable, define the range of image, default draw all the range
        metrics_list: list, used for appointing the metrics for drawing, default='all' 
        '''
        if not epochs_range:
            epochs_range = (1, self.__num_epochs)
        Estimate.__img_process(self, epochs_range, metrics_list=metrics_list)
        plt.tight_layout()
        plt.show()
    
    
    def savefig(self, filename, epochs_range=None, metrics_list='all'):
        '''
        作指标值-迭代次数图像并保存
        
        参数
        ----
        filename：字符串类型，保存的文件名
        epochs_range：列表类型，可选，定义图像的绘制范围，默认全部画出
        metrics_list：列表类型，指定绘图的指标函数列表，默认为全部指标
        
        
        Save the metrics - epochs image
        
        Parameters
        ----------
        filename: str, file name
        epochs_range: list, callable, define the range of image, default draw all the range
        metrics_list: list, used for appointing the metrics for drawing, default='all' 
        '''
        if not epochs_range:
            epochs_range = (1, self.__num_epochs)
        Estimate.__img_process(self, epochs_range, metrics_list=metrics_list)
        plt.tight_layout()
        plt.savefig(filename)



class Search():
    '''
    对keras模型或与keras模型相同架构的模型进行网格搜索
    
    参数
    ----
    data：自变量数据
    targets：目标数据
    verbos：整型，可选，可以是0、1、2，默认为0
    
    属性
    ----
    comb：最佳组合
    
    
    Grid searching for keras model or other similar models
    
    Parameters
    ----------
    data: independent variable
    targets: dependent variable
    verbose: int, callable, can be 0, 1, 2, default = 0
    
    Atrributes
    ----------
    comb: the best combination
    '''
    def __init__(self, data, targets, verbose=0):
        self.data = data
        self.targets = targets
        self.__verbose = verbose
    
    
    def __part_of_kfold(self, num_validation_samples, num_part):
        validation_data = self.data[num_validation_samples * num_part:
                                    num_validation_samples * (num_part + 1)]
        validation_targets = self.targets[num_validation_samples * num_part:
                                          num_validation_samples * (num_part + 1)]
        
        train_data = np.concatenate([self.data[: num_validation_samples * num_part],
                                     self.data[num_validation_samples * (num_part + 1):]])
        train_targets = np.concatenate([self.targets[: num_validation_samples * num_part],
                                        self.targets[num_validation_samples * (num_part + 1):]])
        
        return train_data, train_targets, validation_data, validation_targets
    
    
    def __part_of_gridsearch(self, build_model, num_epochs, batch_size, param, k, n, select_method, verbose, shuffle):
        num_validation_samples = len(self.data) // k
        
        param_keys_list = list(param.keys())
        param_values_list = np.array(list(param.values()))
        param_values_list = np.meshgrid(*param_values_list)
        num_param_values_list = len(param_values_list)
        for i in range(num_param_values_list):
            param_values_list[i] = param_values_list[i].flatten()
        param_values_list = np.array(param_values_list)
        param_values_list = param_values_list.T
        
        val_m = []
        val_m_dict = []
        val_m_loc = []
        for param_values in param_values_list:
            param_dict = dict(zip(param_keys_list, param_values))
            val_list = np.zeros(num_epochs)
            for i in range(n):
                if shuffle:
                    local_data = self.data
                    local_targets = self.targets
                    state = np.random.get_state()
                    np.random.shuffle(local_data)
                    np.random.set_state(state)
                    np.random.shuffle(local_targets)
                for j in range(k):
                    train_data, train_targets,\
                    validation_data, validation_targets = Search.__part_of_kfold(
                        self, num_validation_samples=num_validation_samples, num_part=j)
                    model = build_model(**param_dict)
                    history = model.fit(train_data, train_targets,
                                        validation_data=(validation_data, validation_targets),
                                        epochs=num_epochs, batch_size=batch_size, verbose=verbose)
                    val_list += history.history['val_'+select_method[0]]
                    
            val_list /= (n * k)  #这是网格中一种情况的结果
            if select_method[1] == 'min':
                location_ = np.argmin(val_list)
                val_m.append(val_list[location_])
                val_m_loc.append(location_)
            elif select_method[1] == 'max':
                location_ = np.argmax(val_list)
                val_m.append(val_list[location_])
                val_m_loc.append(location_)
            
            val_m_dict.append(param_dict)
        
        if select_method[1] == 'max':
            final_loc = np.argmin(val_m)
            final_dict = val_m_dict[final_loc]
            final_val_m = val_m[final_loc]
            final_epoch = val_m_loc[final_loc]
            
            return final_dict, final_val_m, final_epoch
            self.collocation = {'collocation': final_dict,
                                'loss': final_val_m,
                                'epoch': final_epoch}
        
        elif select_method[1] == 'min':
            final_loc = np.argmax(val_m)
            final_dict = val_m_dict[final_loc]
            final_val_m = val_m[final_loc]
            final_epoch = val_m_loc[final_loc]
            
            return final_dict, final_val_m, final_epoch
            self.collocation = {'collocation': final_dict,
                                'acc': final_val_m,
                                'epoch': final_epoch}
    
    
    def gridsearch(self, build_model, num_epochs, batch_size, param, k=5, n=1, select_method=('loss', 'min'), shuffle=True, random_state=None):
        '''
        参数
        ----
        buil_model：模型函数或函数列表，需要返回一个model
        num_epochs：整型，迭代次数
        batch_size：整型，批量大小
        param：字典或字典列表类型，表示build_model函数的参数，参数名为键，待选参数值列表为值
        k：整型，可选，k折验证的折叠数，默认为5
        n：整型，可选，k折验证的重复次数，默认为1
        select_method：元组类型，评估方法，形式为('指标', 'max'或'min')，默认为('loss', 'min')
        shuffle：布尔类型，可选，True表示打乱数据，False表示不打乱数据，默认为True
        random_state：整型，可选，随机种子
        
        
        Parameters
        ----------
        build_model: the function of function list to generate keras-model, which should return a model
        num_epochs: int, the number of epochs
        batch_size: int, the size of batch
        param: dict or list of dicts, param needs to be input a dictionary with parm_name as key and param_value (list) as value
        k: int, callable, the number of k-fold, default = 5
        n: int, callable, the number of repetition of k-fold, default = 1
        select_method: tuple, method of selecting the best collocation, default = ('loss', 'min')
        shuffle: bool, callable, 'True' will shuffle the data, 'False' will not, default = True
        random_state: int, callable, random seed
        '''
        np.random.seed(random_state)
        if type(build_model).__name__ == 'function':
            final_dict, final_val_m, final_epoch =\
            Search.__part_of_gridsearch(self, build_model=build_model, num_epochs=num_epochs,\
                                        batch_size=batch_size, param=param,\
                                        k=k, n=n, select_method=select_method,\
                                        verbose=self.__verbose, shuffle=shuffle)
        else:
            last_dict_list = []
            last_val_m_list = []
            last_epoch_list = []
            num_of_models = len(build_model)
            for i in num_of_models:
                last_dict, last_val_m, last_epoch =\
                Search.__part_of_gridsearch(self, build_model=build_model, num_epochs=num_epochs,\
                                            batch_size=batch_size, param=param,\
                                            k=k, n=n, select_method=select_method,\
                                            verbose=self.__verbose, shuffle=shuffle)
                last_dict_list.append(last_dict)
                last_val_m_list.append(last_val_m)
                last_epoch_list.append(last_epoch)
            
            if select_method[1] == 'min':
                last_loc = np.argmin(last_val_m_list)
                final_dict = last_dict_list[last_loc]
                final_val_m = last_val_m_list[last_loc]
                final_epoch = last_epoch_list[last_loc]
            
            elif select_method[1] == 'max':
                last_loc = np.argmax(last_val_m_list)
                final_dict = last_dict_list[last_loc]
                final_val_m = last_val_m_list[last_loc]
                final_epoch = last_epoch_list[last_loc]
        
        self.comb = {'comb': final_dict, select_method[0]: final_val_m, 'epoch': final_epoch}



class Error():
    '''
    误差分析
    
    参数
    ----
    func：函数类型，模型的预测函数
    
    属性
    ----
    abs_error：1-D ndarray数组，绝对误差列表
    rel_error：1-D ndarray数组，相对误差列表
    abs_sort：绝对误差从大到小的排序
    rel_sort：相对误差从大到小的排序
    mae：平均绝对误差
    mape：平均绝对百分比误差
    mse：平均平方误差
    rmse：均方根误差
    sae：绝对误差和
    sse：残差平方和
    
    
    Error Analysis
    
    Parameter
    ---------
    func: function, predicting function of models
    
    Atrributes
    ----------
    abs_error: 1-D ndarray, absolute error list
    rel_error: 1-D ndarray, relative error list
    abs_sort: list of absolute values of errors sorted from large to small
    rel_sort: list of relative values of errors sorted from large to small
    mae: mean absolute error
    mape: mean absolute percentage error
    mse: mean squared error
    rmse: root mean square error
    sae: sum of absolute error
    sse: sum of squared error
    '''
    def __init__(self, func):
        self.__func = func
    
    
    def fit(self, data, target):
        data = np.array(data)
        target = np.array(target)
        self.__num_data = len(target)
        
        predict_target = self.__func(data)
        self.abs_error = abs(target - predict_target)
        self.rel_error = abs((target - predict_target) / target)
    
    @property
    def abs_sort(self):
        error_index = np.arange(self.__num_data)
        self.abs_sort = sorted(list(zip(self.abs_error, error_index)), reverse=True)
    
    @property
    def rel_sort(self):
        error_index = np.arange(self.__num_data)
        self.rel_sort = sorted(list(zip(self.rel_error, error_index)), reverse=True)
    
    @property
    def sse(self):
        return sum(self.abs_error**2)
    
    @property
    def sae(self):
        return sum(self.abs_error)
    
    @property
    def mse(self):
        return sum(self.abs_error**2) / self.__num_data
    
    @property
    def mae(self):
        return sum(self.abs_error) / self.__num_data
    
    @property
    def rmse(self):
        return (sum(self.abs_error**2) / self.__num_data)**0.5
    
    @property
    def mape(self):
        return sum(self.rel_error) / self.__num_data



class Sense():
    '''
    灵敏度分析
    r = (x0, x1, x2, ..., xn)
    y = f(r)
    第i个特征在r=r0时的灵敏度：
    s(xi, r0)= (dy/dxi) * (xi/y)   (r=r0)
    
    参数
    ----
    func：函数类型，模型的预测函数
    x_index：列表类型，可选，需要求灵敏度的特征索引列表，默认为全部
    acc：浮点数类型，可选，求导的精度，默认为0.05
    
    属性
    ----
    s_mat：由特征的灵敏度值组成的矩阵
    prediction：预测值列表
    
    
    Sensitivity Analysis
    r = (x0, x1, x2, ..., xn)
    y = f(r)
    the sensitivity of the ith feature at r=r0:
    s(xi, r0)= (dy/dxi) * (xi/y)   (r=r0)
    
    Parameters
    ----------
    func: function, predicting function of models
    x_index: list, callable, index of features whose sensitivity needs to be calculated, default to all
    acc: float, callable, accuracy of derivation, default=0.05
    
    Attributes
    ----------
    s_mat: matrix combined of sensitivities of features
    prediction: list of predicted values
    '''
    def __init__(self, func, x_index=None, acc=0.05):
        self.__func = func
        self.x_index = x_index
        self.acc = acc
        self.s_mat = []
        self.prediction = []
    
    def fit(self, x0):
        '''
        参数
        ----
        x0：特征的初始值
        
        
        Parameter
        ---------
        x0: initial values of features
        '''
        if not self.x_index:
            self.x_index = range(len(x0))
        
        acc2 = 2 * self.acc
        func0 = self.__func(x0)
        self.prediction.append(func0)
        s_list = []
        
        for i in self.x_index:
            x0[i] += self.acc
            func1 = self.__func(x0)
            x0[i] -= acc2
            func2 = self.__func(x0)
            de = (func1 - func2) / (acc2 * func0) * x0[i]
            x0[i] += self.acc
            s_list.append(de)
        
        self.s_mat.append(s_list)
    
    
    def clr(self):
        '''
        清空s_mat和prediction
        
        
        Clear the s_mat and the prediction
        '''
        self.s_mat = []
        self.prediction = []


'''
本模块是额外工具箱

This module is extra Utils
'''

import numpy as np
from matplotlib import pyplot as plt


def time_before(time_list, time, itself=False, sep=True):
    '''
    寻找特定时间
    在时间列表中寻找某个时间点之前的时间
    要求time参数中时间的表述为'xxxx/xx/xx'，分别对应年-月-日，/表示可以是任意符号
    如“1970年1月1日”应表述为'1970/01/01'或'1970.01.01'等
    
    参数
    ----
    time_list：列表类型，时间列表
    time：字符串类型，设定的时间点
    itself：布尔类型，寻找时是否包括设定的时间点本身
    sep：布尔类型，time_list中的元素是否有符号将年月日分开，默认为True
    
    
    Search special time
    search all then time in time list which are before the point you set
    the expression of time are required to be like 'xxxx/xx/xx', corresponding to year-month-date, / can be any symbol
    e.g.  "January 1st, 1970" should be turned to '1970/01/01' or '1970.01.01' etc
    
    Parameters
    ----------
    time_list: list, list of time
    time: str, the time point
    itself: bool, if include the time point itself when searching
    sep: bool, True means elements in time_list have symbols to seperate the year, month and day, default=True
    '''
    select_time = []
    select_index = []
    date = int(time[:4]+time[5:7]+time[8:10])
    
    for i in range(len(time_list)):
        judge_time = time_list[i]
        if sep:
            judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
        else:
            judge_num = int(judge_time)
        
        if date > judge_num:
            select_time.append(judge_time)
            select_index.append(i)
        
        if itself:
            if date == judge_num:
                select_time.append(judge_time)
                select_index.append(i)
    
    return select_time, select_index


def time_after(time_list, time, itself=False, sep=True):
    '''
    寻找特定时间
    在时间列表中寻找某个时间点之后的时间
    要求时间的表述为'xxxx/xx/xx'，分别对应年-月-日，/表示可以是任意符号
    如“1970年1月1日”应表述为'1970/01/01'或'1970.01.01'等
    
    参数
    ----
    time_list：列表类型，时间列表
    time：字符串类型，设定的时间点
    itself：布尔类型，寻找时是否包括设定的时间点本身
    sep：布尔类型，time_list中的元素是否有符号将年月日分开，默认为True
    
    
    Search special time
    search all then time in time list which are after the point you set
    the expression of time are required to be like 'xxxx/xx/xx', corresponding to year-month-date, / can be any symbol
    e.g.  "January 1st, 1970" should be turned to '1970/01/01' or '1970.01.01' etc
    
    Parameters
    ----------
    time_list: list, list of time
    time: str, the time point
    itself: bool, if include the time point itself when searching
    sep: bool, True means elements in time_list have symbols to seperate the year, month and day, default=True
    '''
    select_time = []
    select_index = []
    date = int(time[:4] + time[5:7] + time[8:10])
    
    for i in range(len(time_list)):
        judge_time = time_list[i]
        if sep:
            judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
        else:
            judge_num = int(judge_time)
        
        if judge_num > date:
            select_time.append(judge_time)
            select_index.append(i)
        
        if itself:
            if date == judge_num:
                select_time.append(judge_time)
                select_index.append(i)
    
    return select_time, select_index


def time_equal(time_list, time, sep=True, equal=True):
    '''
    寻找特定时间的索引
    在时间列表中寻找某个时间点
    要求时间的表述为'xxxx/xx/xx'，分别对应年-月-日，/表示可以是任意符号
    如“1970年1月1日”应表述为'1970/01/01'或'1970.01.01'等
    
    参数
    ----
    time_list：列表类型，时间列表
    time：列表类型，设定的时间点列表
    sep：布尔类型，time_list中的元素是否有符号将年月日分开，默认为True
    equal：布尔类型，True表示搜索与time相同的时间，False表示搜索与time不相同的时间
    
    
    Search index of special time
    search all then time in time list which are the same as the point you set
    the expression of time are required to be like 'xxxx/xx/xx', corresponding to year-month-date, / can be any symbol
    e.g.  "January 1st, 1970" should be turned to '1970/01/01' or '1970.01.01' etc
    
    Parameters
    ----------
    time_list: list, list of time
    time: list, the time point list
    sep: bool, True means elements in time_list have symbols to seperate the year, month and day, default=True
    equal: bool, True means searching the time in "time", False means searching the time not in "time"
    '''
    select_index = []
    select_time = []
    for i in range(len(time)):
        time[i] = int(time[i][:4] + time[i][5:7] + time[i][8:10])
    
    if equal:
        for i in range(len(time_list)):
            judge_time = time_list[i]
            if sep:
                judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
            else:
                judge_num = int(judge_time)
            
            if judge_num in time:
                select_index.append(i)
                select_time.append(judge_time)
    
    else:
        for i in range(len(time_list)):
            judge_time = time_list[i]
            if sep:
                judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
            else:
                judge_num = int(judge_time)
            
            if not judge_num in time:
                select_index.append(i)
                select_time.append(judge_time)
        
        return select_time, select_index


def time_between(time_list, begin, end, begin_itself=False, end_itself=False, sep=True):
    '''
    寻找特定时间
    在时间列表中寻找某个时间点之后的时间
    要求时间的表述为'xxxx/xx/xx'，分别对应年-月-日，/表示可以是任意符号
    如“1970年1月1日”应表述为'1970/01/01'或'1970.01.01'等
    
    参数
    ----
    time_list：列表类型，时间列表
    begin：字符串类型，设定的开始时间点
    end：字符串类型，设定的结束时间点
    begin_itself：布尔类型，寻找时是否包括设定的开始时间点本身
    end_itself：布尔类型，寻找时是否包括设定的结束时间点本身
    sep：布尔类型，time_list中的元素是否有符号将年月日分开，默认为True
    
    
    Search special time
    search all then time in time list which are after the point you set
    the expression of time are required to be like 'xxxx/xx/xx', corresponding to year-month-date, / can be any symbol
    e.g.  "January 1st, 1970" should be turned to '1970/01/01' or '1970.01.01' etc
    
    Parameters
    ----------
    time_list: list, list of time
    begin: str, the beginning time point
    end: str, the end time point
    bengin_itself: bool, if include the beginning time point itself when searching
    end_itself: bool, if include the end time point itself when searching
    sep: bool, True means elements in time_list have symbols to seperate the year, month and day, default=True
    '''
    select_time = []
    select_index = []
    
    # begin
    date_1 = int(begin[:4] + begin[5:7] + begin[8:10])
    
    # end
    date_2 = int(end[:4] + end[5:7] + end[8:10])
    
    if not begin_itself and not end_itself:
        for i in range(len(time_list)):
            judge_time = time_list[i]
            if sep:
                judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
            else:
                judge_num = int(judge_time)
            
            if date_1 < judge_num < date_2:
                select_time.append(judge_time)
                select_index.append(i)
    
    elif begin_itself and not end_itself:
        for i in range(len(time_list)):
            judge_time = time_list[i]
            if sep:
                judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
            else:
                judge_num = int(judge_time)
            
            if date_1 <= judge_num < date_2:
                select_time.append(judge_time)
                select_index.append(i)
    
    elif not begin_itself and end_itself:
        for i in range(len(time_list)):
            judge_time = time_list[i]
            if sep:
                judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
            else:
                judge_num = int(judge_time)
            
            if date_1 < judge_num <= date_2:
                select_time.append(judge_time)
                select_index.append(i)
    
    else:
        for i in range(len(time_list)):
            judge_time = time_list[i]
            if sep:
                judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
            else:
                judge_num = int(judge_time)
            
            if date_1 <= judge_num <= date_2:
                select_time.append(judge_time)
                select_index.append(i)
    
    return select_time, select_index


def grid_net(*args):
    '''
    生成网格点
    将输入的列表遍历组合
    
    
    Generate grid
    traverse and combine the input list
    '''
    net = np.meshgrid(*args)
    for i in range(len(net)):
        net[i] = net[i].flatten()
    net = np.vstack(tuple(net)).T
    return net


class rc():
    '''
    相关系数
    
    
    correlation coefficient
    '''
    def fit(self, *arg):
        arg = np.array(arg, dtype=float)
        if len(arg.shape) != 2:
            raise Exception("Input list should be 1-D.")
        
        cov_mat = np.cov(arg)
        var_mat = np.diagonal(cov_mat)**0.5
        var_mat[var_mat == 0] = 1
        
        for i in range(cov_mat.shape[0]):
            cov_mat[i] /= var_mat[i]
            cov_mat[:, i] /= var_mat[i]
        
        self.rc_mat = cov_mat
    
    
    def show(self, index=None, cmap='Blues'):
        '''
        参数
        ----
        index：列表形式，可选，各数组名称
        cmap：字符串形式，可选，颜色板，默认为'Blues'
        
        
        Parameters
        ----------
        index: list, callable, names of each array
        cmap: str, callable, color board, default='Blues'
        '''
        plt.matshow(self.rc_mat, cmap=cmap)
        plt.colorbar()
        if index:
            n_list = range(len(index))
            plt.xticks(n_list, index)
            plt.yticks(n_list, index)
        plt.show()
    
    
    def savefig(self, filename, index=None, cmap='Blues'):
        '''
        参数
        ----
        filename：字符串形式，文件名
        index：列表形式，可选，各数组名称
        cmap：字符串形式，可选，颜色板，默认为'Blues'
        
        
        Parameters
        ----------
        filename: str, file name
        index: list, callable, names of each array
        cmap: str, callable, color board, default='Blues'
        '''
        plt.matshow(self.rc_mat, cmap=cmap)
        plt.colorbar()
        if index:
            n_list = range(len(self.rc_mat))
            plt.xticks(n_list, index)
            plt.yticks(n_list, index)
        plt.savefig(filename)
    
    

def e_distance(p1, p2):
    '''
    求某两个点之间的欧式距离
    
    参数
    ----
    p1：一维数组，第一个点的位置
    p2：一维数组，第二个点的位置
    
    
    Calculate the Euclidean distance between two points
    
    Parameters
    ----------
    p1: 1-D list, the location of the first point
    p2: 1-D list, the location of the second point
    '''
    p1 = np.array(p1)
    p2 = np.array(p2)
    
    return sum((p1 - p2)**2)**0.5
    
    
def e_distances(data, des='o'):
    '''
    求data到目标点的欧式距离
    
    参数
    ----
    data：一维或二维列表，数据
    des：字符串或一维数组，可选'o'或'O'(原点)、'mean'(均值点)及自定义数组，目标点坐标，默认为'o'
    
    
    Calculate the Euclidean distances between data and destination
    
    Parameter
    ---------
    data: 1-D or 2-D list, data
    des: str or 1-D list, 'o' or 'O' (origin), 'mean' (mean point) and custom array are optional, the coordinate of destination, default='o'
    '''
    data = np.array(data, dtype=np.float)
    if len(data.shape) < 3:
        n = data.shape[-1]
    else:
        raise Exception("Parameter 'data' must be 1-D or 2-D.")

    if des == 'o' or des == 'O':
        des = np.zeros(n)

    elif des == 'mean':
        des = data.mean(axis=0)
        
    else:
        des = np.array(des)
    
    data -= des
    data = data**2
    result = data.sum(axis=len(data.shape)-1)
    result = result**0.5
    
    return result

    
def m_distance(data, p1, p2):
    '''
    求某两个点之间的马氏距离
    
    参数
    ----
    data：二维列表，数据
    p1：一维或二维数组，第一个点的位置
    p2：一维或二维数组，第二个点的位置
    
    
    Calculate the Mahalanobis distance between two points
    
    Parameters
    ----------
    data: 2-D list, data
    p1: 1-D or 2-D list, the location of the first point
    p2: 1-D or 2-D list, the location of the second point
    '''
    data = np.mat(data, dtype=np.float)
    dataT = data.T
    
    if len(data.shape) != 2:
        raise Exception("Parameter 'data' must be 2-D.")
    
    SI = np.mat(np.cov(dataT)).I

    p1 = np.mat(p1, dtype=np.float)
    p2 = np.mat(p2, dtype=np.float)
    result = (p1 - p2) * SI * (p1 - p2).T
    
    return result[0, 0]**0.5


def m_distances(data, des='o'):
    '''
    求data到目标点的马氏距离
    
    参数
    ----
    data：二维列表，数据
    des：字符串或一维或二维数组，可选'o'或'O'(原点)、'mean'(均值点)及自定义数组，目标点坐标，默认为'o'
    
    
    Calculate the Mahalanobis distance between data and destination
    
    Parameter
    ---------
    data: 2-D list, data
    des: str or 1-D or 2-D list, 'o' or 'O' (origin), 'mean' (mean point) and custom array are optional, the coordinate of destination, default='o'
    '''
    data = np.mat(data, dtype=np.float)
    dataT = data.T
    
    if len(data.shape) != 2:
        raise Exception("Parameter 'data' must be 2-D.")
    
    SI = np.mat(np.cov(dataT)).I

    if des == 'o' or des == 'O':
        des = np.zeros(data.shape[-1])
    
    elif des == 'mean':
        des = data.mean(axis=0)
    
    else:
        des = np.mat(des)
    
    return np.diag((data - des) * SI *(dataT - des.T))**0.5


'''
本模块储存着几种常用坐标变换

This module stores several common coordinate transformations
'''

import numpy as np
from maysics import constant


def lorentz(v, x):
    '''
    洛伦兹坐标变换
    
    参数
    ----
    v：惯性系的相对速度，当速度方向沿着x轴正方向时（正变换）v为正
    x：列表，(x, y, z, t)
    
    
    Lorentz coordinate transformation, S ←→ S'
    
    Parameters
    ----------
    v: relative velocity of inertial system, 'v' is positive when the velocity direction is along the positive direction of x-axis (positive transformation)
    x: list, (x, y, z, t)
    '''
    γ = 1 / (1 - (v / constant.c)**2)**0.5
    xp = γ * (x[0] - v * x[3])
    yp = x[1]
    zp = x[2]
    tp = γ * (x[3] - v * x[0] / constant.c**2)
    
    return np.array([xp, yp, zp, tp])
    

def lorentz_v(v, vo):
    '''
    洛伦兹速度变换
    
    参数
    ----
    v：惯性系的相对速度，当速度方向沿着x轴正方向时（正变换）v为正
    vo：列表，(vx, vy, vz)
    
    
    Lorentz speed transformation, S ←→ S'
    
    Parameters
    ----------
    v: relative velocity of inertial system, 'v' is positive when the velocity direction is along the positive direction of x-axis (positive transformation)
    vo: list, (vx, vy, vz)
    '''
    γ = 1 / (1 - (v / constant.c)**2)**0.5
    factor = 1 - v * vo[0] / constant.c**2
    vxp = (vo[0] - v) / factor
    vyp = vo[1] * γ / factor
    vzp = vo[2] * γ / factor
    
    return np.array([vxp, vyp, vzp])


def polar(x):
    '''
    极坐标正变换
    
    参数
    ----
    x：列表，(x, y)
    
    
    Polar positive transformation
    
    Parameters
    ----------
    x: list, (x, y)
    '''
    r = (x[0]**2 + x[1]**2)**0.5
    if x[0] == 0:
        if x[1] > 0:
            theta = np.pi/2
        elif x[1] < 0:
            theta = -np.pi/2
        elif x[1] == 0:
            theta = 0
    else:
        theta = np.arctan(x[1] / x[0])
    
    return np.array([r, theta])


def ipolar(xp):
    '''
    极坐标逆变换
    
    参数
    ----
    xp：列表，(r, θ)
    
    
    Parameters
    ----------
    xp: list, (r, θ)
    '''
    x = xp[0] * np.cos(xp[1])
    y = xp[0] * np.sin(xp[1])
    
    return np.array([x, y])


def cylinder(x):
    '''
    柱坐标正变换
    
    参数
    ----
    x：列表，(x, y, z)
    
    
    Cylinder positive transformation
    
    Parameters
    ----------
    x: list, (x, y, z)
    '''
    r = (x[0]**2 + x[1]**2)**0.5
    if x[0] == 0:
        if x[1] > 0:
            theta = np.pi/2
        elif x[1] < 0:
            theta = -np.pi/2
        elif x[1] == 0:
            theta = 0
    else:
        theta = np.arctan(x[1] / x[0])
    z = x[2]
    
    return np.array([r, theta, z])


def icylinder(xp):
    '''
    柱坐标逆变换
    
    参数
    ----
    xp：列表，(r, θ, z)
    
    
    Cylinder inverse transformation
    
    Parameters
    ----------
    xp: list, (r, θ, z)
    '''
    x = xp[0] * np.cos(xp[1])
    y = xp[0] * np.sin(xp[1])
    z = xp[2]
    
    return np.array([x, y, z])


def sphere(x):
    '''
    球坐标正变换
    
    参数
    ----
    x：列表，(x, y, z)
    
    
    Sphere positive transformation
    
    Parameters
    ----------
    x: list, (x, y, z)
    '''
    r = (x[0]**2 + x[1]**2 + x[2]**2)**0.5
    if r == 0:
        theta = 0
        phai = 0
    else:
        phai = np.arccos(x[2] / r)
        if x[0] == 0:
            if x[1] > 0:
                theta = np.pi/2
            elif x[1] < 0:
                theta = -np.pi/2
            elif x[1] == 0:
                theta = 0
        else:
            theta = np.arctan(x[1] / x[0])
    
    return np.array([r, theta, phai])


def isphere(xp):
    '''
    球坐标逆变换
    
    参数
    ----
    xp：列表，(r, θ, φ)
    
    
    Sphere inverse transformation
    
    Parameters
    ----------
    xp: list, (r, θ, φ)
    '''
    x = xp[0] * np.sin(xp[2]) * np.cos(xp[1])
    y = xp[0] * np.sin(xp[2]) * np.sin(xp[1])
    z = xp[0] * np.cos(xp[2])
    
    return np.array([x, y, z])


def rotate(theta, x):
    '''
    平面直角坐标系的旋转变换
    逆时针旋转时theta为正，顺时针旋转时theta为负
    
    参数
    ----
    x：列表，(x, y)
    theta：浮点数类型，坐标系绕原点逆时针旋转的角度
    
    
    Rotation transformation of plane rectangular coordinate system
    'theta' is positive when rotating anticlockwise and negative when rotating clockwise
    
    Parameter
    ---------
    x: list, (x, y)
    theta: float, the angle that the coordinate system rotates counterclockwise about the origin
    '''
    xp = np.cos(theta) * x[0] + np.sin(theta) * x[1]
    yp = np.cos(theta) * x[1] - np.sin(theta) * x[0]
    
    return np.array([xp, yp])


'''
本模块储存着常用算符和积分方法
调用模块中的算符时应注意：
1、算符作用的多元函数的自变量需为列表(list、tuple、ndarray均可，下同)
2、算符作用的标量函数的输出值需为单个值
3、算符作用的矢量函数的输出值需为列表

This module stores the common operators and integration method
When calling operators in the module, pay attention to:
1. The argument of the multivariate function acted by the operators should be list (list, tuple, ndarray, the same below)
2. The output value of scalar function acted by the operators should be a single value
3. The output value of vector function acted by the operators should be list
'''

import numpy as np
from scipy import constants as C
from maysics.utils import grid_net


def lim(func, x0, acc=0.01, method='both'):
    '''
    求极限
    lim x→x0+ f(x) ≈ f(x+dx)
    lim x→x0- f(x) ≈ f(x-dx)
    lim x→x0 ≈ (f(x+dx) + f(x-dx)) / 2
    
    参数
    ----
    func：函数类型
    x0：取极限的点
    acc：浮点数类型，可选，极限精度，即dx，默认为0.01
    method：字符串类型，求极限方法，可选'both'、'rigth'或'+'(右极限)、'left'或'-'(左极限)
    
    
    Calculate the limit values
    lim x→x0+ f(x) ≈ f(x+dx)
    lim x→x0- f(x) ≈ f(x-dx)
    lim x→x0 ≈ (f(x+dx) + f(x-dx)) / 2
    
    Parameters
    ----------
    func: function
    x0: point of limit
    acc: float, callable, the accuracy of calculation, equals to dx, default=0.01
    method: str, the method of calculation, 'both', 'right' or '+'(right limit), 'left' or '-'(left limit) are optional
    '''
    if method == 'both':
        func1 = func(x0 + acc)
        func2 = func(x0 - acc)
        return 0.5 * (func1 + func2)
    
    elif method == '+' or method == 'right':
        return func(x0 + acc)
    
    elif method == '-' or method == 'left':
        return func(x0 - acc)
    
    else:
        raise Exception("Parameter 'method' must be one of 'both', 'right', '+', 'left', '-'.")



def dif(f, dim=0, acc=0.05):
    '''
    对多元函数的求导算子，对矢量函数则相当于对每个分量函数都进行了相同的求导
    多元矢量函数可以分解为多个标量函数
    无法将二维列表识别成一行作为一个自变量，需要循环
    
    参数
    ----
    f:函数类型, 函数的自变量x是列表
    dim：整型，可选，对第dim个自变量求导，默认为0
    acc：浮点数类型，可选，求导的精度，默认为0.05
    
    返回值：一个导函数
    
    
    derivative operator for multivariate function( same derivation to each component)
    Multivariate vector functions can be decomposed into several scalar functions
    Unable to recognize the two-dimensional list as an independent variable, loop is required
    
    Parameters
    ----------
    f: function, argument x of function should be a list
    dim: int, callable, derivation of the dim independent variable, default=0
    acc: float, callable, accuracy of derivation, default=0.05
    
    return: a derivative function
    '''
    def obj(x):
        x = np.array(x, dtype=float)
        x[dim] += acc
        func1 = f(x)
        if type(func1).__name__ == 'tuple' or 'list':
            func1 = np.array(func1)
        x[dim] -= 2 * acc
        func2 = f(x)
        if type(func2).__name__ == 'tuple' or 'list':
            func2 = np.array(func2)
        de = (func1 - func2) / (acc + acc)
        return de
    return obj



def H(f, m, U, acc=0.05):
    '''
    哈密顿算符：H = - hr**2 / 2m * ▽**2 + U
    
    参数
    ----
    f：函数
    m：数，例子质量
    U：数或函数，势能
    acc：浮点数类型，可选，求导的精度，默认为0.05
    
    
    Hamilton: H = - hr**2 / 2m * ▽**2 + U
    
    Parameters
    ----------
    m: number, the mass of the particle
    U: number or function, potential energy
    acc: float, callable, accuracy of derivation, default=0.05
    f: function
    '''
    def obj(x):
        x = np.array(x, dtype=float)
        result = 0
        for i in range(len(x)):
            func1 = 2 * f(x)
            x[i] += 2 * acc
            func2 = f(x)
            x[i] -= 4 * acc
            func3 = f(x)
            x[i] += 2 * acc
            de = (func2 + func3 - 2 * func1) / 4 / acc**2
            result += de
        result *= (C.h / (2 * np.pi))**2 / (2 * m)
        
        if type(U).__name__ == 'function':
            result += U(x) * f(x)
        else:
            result += U * f(x)
        return result
    return obj



class Del():
    '''
    ▽算子
    被作用函数的自变量x是列表
    
    参数
    ----
    acc：浮点数类型，可选，求导的精度，默认为0.05
    
    
    nabla operator ▽
    the argument x of the affected function should be a list
    
    Parameters
    ----------
    acc: float, callable, accuracy of derivation, default=0.05
    '''
    def __init__(self, acc=0.05):
        self.acc = acc
    

    def grad(self, f, dim=None):
        '''
        求标量函数梯度：▽f(r)
        
        参数
        ----
        f：函数，要求函数f返回一个数值
        dim：整型，求导的维度，默认全部求导
        
        返回值：一个梯度函数
        
        
        gradient of scalar function: ▽f(r)
        
        Parameters
        ----------
        f: function, the function should return a number
        dim: int, dimensions for derivation, default: all
        
        return: a gradient function
        '''
        def obj(x):
            x = np.array(x, dtype=float)
            if not dim:
                dim = len(x)
            result = []
            for i in range(dim):
                x[i] += self.acc
                func1 = f(x)
                x[i] -= 2 * self.acc
                func2 = f(x)
                x[i] += self.acc
                de = (func1 - func2) / (self.acc + self.acc)
                result.append(de)
            result = np.array(result)
            return result
        return obj
    

    def dot(self, f):
        '''
        ▽点乘矢量函数：▽·f(r)
        
        参数
        ----
        f：函数，要求函数f返回一个列表

        返回值：一个新函数
        
        
        dot product between ▽ and vector function: ▽·f(r)
        
        Parameter
        ---------
        f: function, the function should return a list
        
        return: a new function
        '''
        def obj(x):
            x = np.array(x, dtype=float)
            result = []
            for i in range(len(x)):
                x[i] += self.acc
                func1 = f(x)[i]
                x[i] -= 2*self.acc
                func2 = f(x)[i]
                x[i] += self.acc
                de = (func1 - func2) / (self.acc + self.acc)
                result.append(de)
            result = np.array(result)
            return result
        return obj
    

    def __diff2forc(self, a, b, f, x):
        '''
        一个中介函数，方便后文程序使用
        '''
        x[b] += self.acc
        func1 = f(x)[a]
        x[b] -= 2*self.acc
        func2 = f(x)[a]
        x[b] += self.acc
        de = (func1 - func2) / (self.acc + self.acc)
        return de
    
    
    def cross(self, f):
        '''
        ▽叉乘矢量函数：▽×f(r)
        
        参数
        ----
        f：函数，要求函数是三维矢量函数

        返回值：一个新函数
        
        
        ▽ cross vector function: ▽×f(r)
        
        Parameter
        ---------
        f: function, the function should be a three-dimension vector function
        
        return: a new function
        '''
        def obj(x):
            x = np.array(x, dtype=float)
            result = np.array([
                Del.__diff2forc(self, 2, 1, f, x) -\
                Del.__diff2forc(self, 1, 2, f, x),
                Del.__diff2forc(self, 0, 2, f, x) -\
                Del.__diff2forc(self, 2, 0, f, x),
                Del.__diff2forc(self, 1, 0, f, x) -\
                Del.__diff2forc(self, 0, 1, f, x)])
            return result
        return obj


    def laplace(self, f):
        '''
        △算子 = ▽**2
        被作用函数的自变量x是列表
        
        参数
        ----
        f：函数
        
        返回值：一个新函数
        
        
        Laplace operator: △ = ▽**2
        the argument x of the affected function should be a list
        
        Parameter
        ---------
        f: function
        
        return: a new function
        '''
        def obj(x):
            x = np.array(x, dtype=float)
            result = 0
            for i in range(len(x)):
                func1 = 2 * f(x)
                x[i] += 2 * self.acc
                func2 = f(x)
                x[i] -= 4 * self.acc
                func3 = f(x)
                x[i] += 2 * self.acc
                de = (func2 + func3 - 2 * func1) / 4 / self.acc**2
                result += de
            return result
        return obj



class inte():
    def __init__(self, condition=None, method='rect', dim=1, random_state=None):
        '''
        定积分
        
        矩形法rect：
        S ≈ ∑ f(xi)dv
        dv = ∏ dx ≈ ∏ Δx = ∏ acc
        
        蒙特卡洛法mc：
        在区域V = area * height中产生loop个随机点
        计算符合条件的随机点个数effect_points
        S ≈ V * effect_points / loop
        
        参数
        ----
        condition：函数类型，可选，条件函数，符合条件的输出Ture，否则输出False，条件函数的输入须为1维数组
        method：字符串类型，可选'rect'、'mc'，默认为'rect'
        dim：整型，1表示被积函数的输入为1维数组，适用于普通自定义函数，2表示被积函数的输入为2维数组，适用于机器学习模型，默认为1
        
        属性
        ----
        history：数组类型，储存每一次的积分结果
        
        
        Definite integral
        
        ractangle method ('rect')：
        S ≈ ∑ f(xi)dv
        dv = ∏ dx ≈ ∏ Δx = ∏ acc
        
        MonteCarlo method ('mc')：
        generate 'loop' random points in area 'V = area * height'
        calculate the number of qualified points 'effect_points'
        S ≈ V * effect_points / loop
        
        Parameters
        ----------
        condition: function, callable, condition function with input as 1-D list, if input if qualified, ouput True, otherwise output False
        method: str, callable, 'rect' and 'mc' are optional, default='rect'
        dim: int, 1 means the input of integrand is 1-D list, like normal functions, 2 means the input of integrand is 2-D list, like machine learning models, default=1
        
        Attribute
        ---------
        history: list, stores every integral result
        '''
        self.__condition = condition
        self.method = method
        self.__dim = dim
        self.history = []
        
        if dim > 2:
            raise Exception("Parameter 'dim' must be one of 1 and 2.")
        
        if method != 'rect' and method != 'mc':
            raise Exception("Parameter 'method' must be one of 'rect' and 'mc'.")
    
    
    def __mc_fit(self, func, area, loop, height, random_state):
        '''
        蒙特卡洛法
        
        
        MonteCarlo method
        '''
        np.random.seed(random_state)
        area = np.array(area)
        area_length = area[:, 1] - area[:, 0]
        V = area_length.prod() * height
        
        num = len(area_length)
        area_points = np.random.rand(loop, num) * area_length + area[:, 0]
        func_points = np.random.rand(loop) * height
        
        effect_points = 0
        if self.__dim == 1:
            for i in range(loop):
                if func_points[i] <= func(area_points[i]) and self.__condition(area_points[i]):
                    effect_points += 1
        
        elif self.__dim == 2:
            func_points_2 = func(area_points)
            for i in range(loop):
                if func_points[i] <= func_points_2[i] and self.__condition(area_points[i]):
                    effect_points += 1
        
        effect_points_rate = effect_points / loop
        
        return V * effect_points_rate
    
    
    def __rect_fit(self, func, area, acc):
        '''
        矩形法
        
        
        Rectangle method
        '''
        acc = np.array(acc)
        dv = acc.prod()
        
        points_net = []
        for i in range(len(area)):
            points_net.append(np.arange(area[i][0], area[i][1], acc[i]))
        points_net = grid_net(*points_net)
        
        if self.__dim == 1:
            func_points=[]
            for i in points_net:
                if self.__condition(i):
                    func_points.append(func(i))
            func_points = np.array(func_points) * dv
        
        elif self.__dim == 2:
            points_net = points_net.tolist()
            for i in points_net[:]:
                if not self.__condition(i):
                    points_net.remove(i)
            points_net = np.array(points_net)
            func_points = func(points_net) * dv
        
        return func_points.sum()
    
    
    def fit(self, func, area,  acc=0.1, loop=10000, height=1, random_state=None):
        '''
        积分
        
        参数
        ----
        func：函数类型，被积函数
        area：二维列表，积分区域
        acc：浮点数或列表类型，可选，积分精度，只有method='rect'时才有效，默认为0.1
        loop：整型，可选，产生的随机数数目，只有method='mc'时才有效，默认为10000
        height：浮点数类型，可选，高度，只有method='mc'时才有效，默认为1
        random_state：整型，随机种子，只有method='mc'时才有效
        
        
        Intergration
        
        Parameters
        ----------
        func: function, integrand
        area: 2-D list, integral region
        acc: float or list, callable, integration accuracy, it's effective only when method='rect', default=0.1
        loop: int, callable, the number of generated random numbers, it's effective only when method='mc', default=10000
        height: float, callable, height, it's effective only when method='mc', default=1
        random_state: int, random seed, it's effective only when method='mc'
        '''
        if self.method == 'rect':
            if type(acc).__name__ == 'float' or type(acc).__name__ == 'int':
                acc = np.ones(len(area)) * acc
            self.history.append(inte.__rect_fit(self, func=func, area=area, acc=acc))
        
        elif self.method == 'mc':
            self.history.append(inte.__mc_fit(self, func=func, area=area, 
                                              loop=loop, height=height, random_state=random_state))
    
    
    def clr(self):
        '''
        清空history
        
        
        Clear the history
        '''
        self.history = []